

class RedisSession:
    def __init__(self):
        self.current_db: int = 0
        self.reader = None
        self.write = None
        self.need_auth_event = None
        self.server = None


__all__ = ["RedisSession", ]



from typing import *
import os
import time
import math
import random
import string
import asyncio
from jinja2 import Template
from .result import *
from .utils import *
from .redis_protocol import RedisProtocol
from .redis_task_node import RedisTaskNode
from .redis_session import RedisSession


class RedisServerBase(RedisProtocol):
    IDLE_TIMEOUT = 60
    WORKER_QUEUE = asyncio.Queue()
    ENABLE_AUTH = False
    MAX_DB_COUNT = 1

    INFO_TEMPLATE = Template(
        """# Server
redis_version:3.0.0
porkpepper_version:0.1.0
porkpepper_mode:{{ porkpepper_mode }}
process_id:{{ process_id }}
run_id:{{ run_id }}
tcp_port:{{ tcp_port }}
uptime_in_seconds:{{ uptime_in_seconds }}
uptime_in_days:{{ uptime_in_days }}

# Clients
connected_clients:{{ connected_clients }}

# Stats
total_connections_received:{{ total_connections_received }}
total_commands_processed:{{ total_commands_processed }}
instantaneous_ops_per_sec:{{ instantaneous_ops_per_sec }}
total_net_input_bytes:{{ total_net_input_bytes }}
total_net_output_bytes:{{ total_net_output_bytes }}

# Replication
role:master
connected_slaves:0
master_repl_offset:0

# Cluster
cluster_enabled:0

# Keyspace
{% for db, info in keyspace %}db{{ db }}:keys={{ info["keys"] }},expires=0,avg_ttl=0\r\n{% endfor %}
""", newline_sequence="\r\n")

    def __init__(self, app=None):
        self._http_app = app
        self.host = None
        self.port = None
        self.connections = set()
        self.start_time = 0
        self.run_id = ""
        self.total_connections_received = 0
        self.total_commands_processed = 0
        self.instantaneous_ops_per_sec = 0
        self.total_net_input_bytes = 0
        self.total_net_output_bytes = 0
        self.init_property()

    def init_property(self):
        self.connections = set()
        self.start_time = int(time.time())
        random_key = ''.join(random.choices(string.ascii_letters, k=128))
        run_id = create_base58_key(random_key, length=20, prefix="PP", timestamp=True)
        self.run_id = run_id
        self.total_connections_received = 0
        self.total_commands_processed = 0
        self.instantaneous_ops_per_sec = 0
        self.total_net_input_bytes = 0
        self.total_net_output_bytes = 0

    @property
    def app(self):
        return self._http_app

    @property
    def password(self):
        return None

    async def handle_stream(self, reader, writer):
        session = None
        try:
            self.connections.add((reader, writer, ))
            self.total_connections_received += 1
            auth = self.ENABLE_AUTH
            need_auth = True if auth else False
            need_auth_event = asyncio.Event()
            if need_auth:
                need_auth_event.set()
            session = RedisSession()
            session.reader = reader
            session.write = writer
            session.need_auth_event = need_auth_event
            session.server = self
            idle_timeout = self.IDLE_TIMEOUT
            while True:
                node_result = await RedisTaskNode.create(session, self)
                if node_result.is_error:
                    raise node_result.error
                if node_result.is_some:
                    node: RedisTaskNode = node_result.unwrap()
                    if node.fa:
                        func, kwargs = node.fa
                        if kwargs:
                            result = await func(**kwargs)
                        else:
                            result = await func()
                        node.result = result
                    write_result = await asyncio.wait_for(node.write_result(session, self),
                                                          timeout=idle_timeout)
                self.total_commands_processed += 1
        except asyncio.TimeoutError as e:
            return Result(e)
        except Exception as e:
            return Result(e)
        finally:
            try:
                writer.close()
                await writer.wait_closed()
            except Exception as e:
                pass
            self.connections.remove((reader, writer, ))
            if session:
                session.reader = None
                session.write = None
                session.server = None

    def info_arguments(self):
        return {
            "porkpepper_version": "0.1.0",
            "porkpepper_mode": "",
            "process_id": os.getpid(),
            "run_id": self.run_id,
            "tcp_port": self.port,
            "uptime_in_seconds": int(time.time()) - self.start_time,
            "uptime_in_days": math.ceil((time.time() - self.start_time) / (24 * 60 * 60)),
            "connected_clients": len(self.connections),
            "keyspace": list(),
            "total_connections_received": self.total_connections_received,
            "total_commands_processed": self.total_commands_processed,
            "instantaneous_ops_per_sec": self.instantaneous_ops_per_sec,
            "total_net_input_bytes": self.total_net_input_bytes,
            "total_net_output_bytes": self.total_net_output_bytes,
        }

    async def serve(self, host='127.0.0.1', port=6379, forever=True, after_start=None, **kwargs):
        self.host = host
        self.port = port
        server = await asyncio.start_server(self.handle_stream, host, port, **kwargs)
        if after_start:
            after_start()
        if forever:
            async with server:
                await server.serve_forever()
        else:
            return server

    async def get(self, session, key) -> Result[bytes]:
        return Result(NotImplementedError())

    async def getset(self, session, key, value) -> Result:
        return Result(NotImplementedError())

    async def set(self, session, key, value) -> Result:
        return Result(NotImplementedError())

    async def key_type(self, session, key):
        return Result(NotImplementedError())

    async def ping(self, session):
        return Result(NotImplementedError())

    async def info(self, session):
        return Result(NotImplementedError())

    async def auth(self, session, password) -> Result[bool]:
        """
        如果是连接状态问题或者非密码对错的情况下，这里应返回 Result[Exception]
        判断密码正确时，应返回 Result[bool]
        :param session:
        :param password:
        :return:
        """
        return Result(NotImplementedError())

    async def ttl(self, session, key):
        return Result(NotImplementedError())

    async def scan(self, session, pattern):
        return Result(NotImplementedError())

    async def dbsize(self, session):
        return Result(NotImplementedError())

    async def select(self, session, db):
        return Result(NotImplementedError())

    async def config(self, session, get_set, field, value=None):
        return Result(NotImplementedError())

    async def delete(self, session, key):
        return Result(NotImplementedError())


__all__ = ["RedisServerBase", ]



import asyncio
from queue import Queue


class FifoLock:
    def __init__(self, maxsize: int = 0):
        self._lock_queue = Queue(maxsize=maxsize)

    async def acquire(self):
        lock = asyncio.Lock()
        if not self._lock_queue.empty():
            await lock.acquire()
        self._lock_queue.put(lock)
        await lock.acquire()

    async def release(self):
        if self._lock_queue.empty():
            return
        self._lock_queue.get()
        if self._lock_queue.empty():
            return
        lock = self._lock_queue.queue[0]
        if lock.locked():
            lock.release()

    async def __aenter__(self):
        await self.acquire()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.release()


__all__ = ["FifoLock", ]



from typing import *
import asyncio
from collections import defaultdict
import aiohttp
from aiohttp import web
from .websocket_session import WebsocketSession


class WebsocketApp(web.Application):
    def __init__(self, session_class=WebsocketSession, **kwargs):
        super(WebsocketApp, self).__init__(**kwargs)
        self._session_class = session_class
        self._session_dict: Dict[str, WebsocketSession] = dict()
        self._user_dict: Dict[str, Set[WebsocketSession]] = defaultdict(set)

    def all_sessions(self):
        return list(self._session_dict.values())

    def all_users(self):
        return list(self._user_dict.keys())

    @property
    def sessions_count(self) -> int:
        return len(self._session_dict)

    @property
    def users_count(self) -> int:
        return len(self._user_dict)

    def get_session(self, session_id: str):
        return self._session_dict.get(session_id, None)

    def get_user(self, user: str) -> Set[WebsocketSession]:
        return self._user_dict.get(user, set())

    def remove_user(self, user, session):
        if user is not None and user in self._user_dict and session in self._user_dict[user]:
            self._user_dict[user].remove(session)
            if not len(self._user_dict[user]):
                self._user_dict.pop(user)

    def add_user(self, user, session):
        if user is not None:
            self._user_dict[user].add(session)

    def remove_session(self, session_id):
        if session_id in self._session_dict:
            self._session_dict.pop(session_id)
            return True
        return False

    def add_session(self, session_id, session):
        self._session_dict[session_id] = session

    def setup_session(self, session, ws):
        session.socket = ws
        session.app = self
        self.add_session(session.session_id, session)

    def cleanup_session(self, session):
        if session.current_user:
            self.remove_user(session.current_user, session)
        self.remove_session(session.session_id)
        session.app = None
        session.socket = None

    async def handler(self, request):
        if not self._session_class:
            return
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        session = self._session_class()
        self.setup_session(session, ws)
        try:
            await session.prepare()
            while True:
                read_timeout = session.read_timeout()
                msg = await ws.receive(read_timeout)
                if msg.type in (aiohttp.WSMsgType.TEXT, msg.type == aiohttp.WSMsgType.BINARY, ):
                    message_result = session.message_loads(msg.data)
                    if message_result.is_some:
                        await session.request(message_result.unwrap())
                elif msg.type == aiohttp.WSMsgType.PING:
                    await ws.pong()
                elif msg.type in (
                        aiohttp.WSMsgType.CLOSE,
                        aiohttp.WSMsgType.CLOSING,
                        aiohttp.WSMsgType.CLOSED,
                        aiohttp.WSMsgType.ERROR,):
                    break
        except Exception as e:
            pass
        finally:
            try:
                await session.on_finish()
            except Exception as e:
                pass
            try:
                await ws.close()
            except Exception as e:
                pass
        self.cleanup_session(session)
        return ws


__all__ = ["WebsocketApp", ]





class UnwrapError(Exception):
    pass


class RedisProtocolFormatError(Exception):
    pass


class ExceptionWithReplyError(Exception):
    def __init__(self, reply="ERR", *args):
        super().__init__(*args)
        self.reply = reply


class NoPasswordError(ExceptionWithReplyError):
    pass


class WrongCommand(Exception):
    pass


class KeyNotFound(Exception):
    pass


class NodeNotFound(Exception):
    pass


class DatabaseNotFound(Exception):
    pass


class CommandNotFound(ExceptionWithReplyError):
    def __init__(self):
        super(CommandNotFound, self).__init__(reply="COMMAND NOT FOUND")


__all__ = [
    "UnwrapError",
    "RedisProtocolFormatError",
    "ExceptionWithReplyError",
    "NoPasswordError",
    "WrongCommand",
    "KeyNotFound",
    "NodeNotFound",
    "DatabaseNotFound",
    "CommandNotFound",
]



import asyncio
import aioredis
from .result import Result
from .websocket_session import WebsocketSession
from .design import WebsocketNode
from .redis_server import RedisServer


class MonitorRedisServer(RedisServer):
    MAX_DB_COUNT = 0

    def info_arguments(self):
        arguments = super(MonitorRedisServer, self).info_arguments()
        arguments.update(porkpepper_mode="monitor")
        return arguments

    async def info(self, session):
        info_arguments = self.info_arguments()
        info = self.INFO_TEMPLATE.render(info_arguments)
        return Result(info)


class SimpleMonitorNode(WebsocketNode):
    MONITOR_TASK_LIST = list()
    TIMER_INTERVAL = 5
    NODE_URL_LIST = list()

    def __init__(self, websocket_path="/monitor"):
        super(SimpleMonitorNode, self).__init__(
            session_class=WebsocketSession,
            websocket_path=websocket_path,
            redis_server=MonitorRedisServer
        )

    async def monitor_task(self, node_url):
        while True:
            connectivity = False
            conn = None
            info = None
            try:
                conn = await aioredis.create_redis(node_url, timeout=1)
                info = await conn.info()
                connectivity = True
            except asyncio.CancelledError:
                return
            except Exception as e:
                connectivity = False
            finally:
                if conn is not None:
                    conn.close()
            message = {
                    "type": "monitorNodeStatus",
                    "nodeUrl": node_url,
                    "connectivity": connectivity,
                    "info": info,
                }
            app = self._http_app
            if app is not None:
                all_sessions = self._http_app.all_sessions()
                send_list = [session.send(message) for session in all_sessions]
                if send_list:
                    asyncio.ensure_future(asyncio.gather(*send_list))
            await asyncio.sleep(self.TIMER_INTERVAL)

    def cancel_monitor_tasks(self):
        task_list = self.MONITOR_TASK_LIST
        for task in task_list:
            task.cancel()
        task_list.clear()

    async def on_start(self):
        self.cancel_monitor_tasks()
        monitor_tasks = [asyncio.create_task(self.monitor_task(node_url)) for node_url in self.NODE_URL_LIST]
        self.MONITOR_TASK_LIST.extend(monitor_tasks)

    async def on_shutdown(self):
        self.cancel_monitor_tasks()


__all__ = ["SimpleMonitorNode", ]



from typing import *
import fnmatch
import json
import inspect
import asyncio
from .error import *
from .result import Result
from .node import PorkPepperNode
from .redis_server import RedisServer


PORKPEPPER_ATTR = "__porkpepper__"


class SocketBasedRedisServer(RedisServer):
    MAX_DB_COUNT = 2
    LOADS = json.loads

    def _session_db_size(self) -> int:
        app = self.app
        if app is not None:
            return app.sessions_count
        else:
            return 0

    def _user_db_size(self) -> int:
        app = self.app
        if app is not None:
            return app.users_count
        else:
            return 0

    @classmethod
    def _session_to_dict(cls, session):
        return dict(
                    type="session",
                    session_id=session.session_id,
                    create_timestamp=session.create_timestamp,
                    current_user=session.current_user
                )

    async def get(self, session, key) -> Result[bytes]:
        app = self.app
        if session.current_db == 0:
            session = app.get_session(key) if app is not None else None
            if session:
                status = self._session_to_dict(session)
                return Result(json.dumps(status))
        elif session.current_db == 1:
            sessions = app.get_user(key) if app is not None else None
            if sessions:
                status = dict(type="user", sessions=[self._session_to_dict(session) for session in sessions])
                return Result(json.dumps(status))
        return Result(None)

    async def getset(self, session, key, value) -> Result:
        return Result(NotImplementedError())

    @classmethod
    async def send_task(cls, session, message):
        try:
            await session.send(message)
        except Exception as e:
            pass

    async def set(self, session, key, value) -> Result:
        app = self.app
        if session.current_db == 0:
            session = app.get_session(key) if app is not None else None
            if session:
                asyncio.ensure_future(asyncio.Task(self.send_task(session, self.LOADS(value))))
        return Result(True)

    async def key_type(self, session, key):
        return Result("string")

    async def ping(self, session):
        return Result("PONG")

    def info_arguments(self):
        service_list = list()
        service_list.append(
            (0,
             dict(
                 keys=self._session_db_size(),
             )
             ))
        service_list.append(
            (1,
             dict(
                 keys=self._user_db_size(),
             )
             ))
        arguments = super(SocketBasedRedisServer, self).info_arguments()
        arguments.update(porkpepper_mode="websocket", keyspace=service_list)
        return arguments

    async def info(self, session):
        info_arguments = self.info_arguments()
        info = self.INFO_TEMPLATE.render(info_arguments)
        return Result(info)

    async def auth(self, session, password) -> Result[bool]:
        if self.password is None:
            return Result(NoPasswordError("ERR Client sent AUTH, but no password is set"))
        auth = self.password or ""
        return Result(password == auth.encode("utf8"))

    async def ttl(self, session, key):
        return Result(-1)

    def _scan_sessions(self, pattern=None):
        app = self.app
        keys = list()
        sessions = app.all_sessions() if app is not None else list()
        for session in sessions:
            if pattern is not None:
                if fnmatch.fnmatch(session.session_id, pattern):
                    keys.append(session.session_id)
            else:
                keys.append(session.session_id)
        return keys

    def _scan_users(self, pattern=None):
        app = self.app
        keys = list()
        users = app.all_users() if app is not None else list()
        for user in users:
            if pattern is not None:
                if fnmatch.fnmatch(user, pattern):
                    keys.append(user)
            else:
                keys.append(user)
        return keys

    async def scan(self, session, pattern):
        if session.current_db == 0:
            keys = self._scan_sessions(pattern)
            return Result(keys)
        if session.current_db == 1:
            keys = self._scan_users(pattern)
            return Result(keys)
        return Result(list())

    async def dbsize(self, session):
        if session.current_db == 0:
            size = self._session_db_size()
            return Result(size)
        if session.current_db == 1:
            size = self._user_db_size()
            return Result(size)
        return Result(0)

    async def select(self, session, db):
        if 0 <= db < self.MAX_DB_COUNT:
            session.current_db = db
            return Result(True)
        else:
            return Result(DatabaseNotFound())

    async def config(self, session, get_set, field, value=None):
        if get_set == b'GET' and field == b'databases':
            return Result([b'databases', f'{self.MAX_DB_COUNT}'.encode("utf8")])
        return Result(CommandNotFound())

    async def delete(self, session, key):
        if session.current_db == 0:
            app = self.app
            if app is not None:
                session = app.get_session(key)
                if session:
                    await session.close()
                return Result(1 if session else 0)
        return Result(0)


class WebsocketNode(PorkPepperNode):
    def __init__(self, session_class, websocket_path, redis_server=SocketBasedRedisServer, **kwargs):
        super(WebsocketNode, self).__init__(
            redis_server=redis_server,
            session_class=session_class,
            websocket_path=websocket_path,
            **kwargs
        )

    async def start(self, service_map: Dict[int, Type] = None, redis_host="127.0.0.1", redis_port=6379, **kwargs):
        return await super(WebsocketNode, self).start(
            enable_websocket=True,
            redis_host=redis_host,
            redis_port=redis_port, **kwargs
        )

    async def serve(self, service_map: Dict[int, Type] = None, redis_host="127.0.0.1", redis_port=6379, **kwargs):
        await super(WebsocketNode, self).serve(
            enable_websocket=True,
            redis_host=redis_host,
            redis_port=redis_port,
            **kwargs
        )


def service(key: str, output: bool = False, description: Optional[str] = None, meta: Optional[Dict] = None):
    def wrap(func):
        signature = inspect.signature(func)
        setattr(func, PORKPEPPER_ATTR, dict(
            key=key,
            output=output,
            signature=signature,
            description=description,
            meta=meta if meta else dict()
        ))
        return func

    return wrap


class ServiceBasedRedisServer(RedisServer):
    MAX_DB_COUNT = 0
    SERVICE_MAP = dict()

    async def get(self, session, key) -> Result[bytes]:
        if session.current_db < 0 or session.current_db >= self.MAX_DB_COUNT:
            return Result(DatabaseNotFound())
        current_service = self.SERVICE_MAP.get(session.current_db, None)
        if not current_service:
            return Result(DatabaseNotFound())
        service_map = current_service["map"]
        service_dict = service_map.get(key, None)
        if not service_dict:
            return Result(None)
        service_arguments = service_dict["args"]
        output = {
            "type": "api",
            "key": service_arguments["key"],
            "output": service_arguments["output"],
            "signature": str(service_arguments["signature"]),
            "description": service_arguments["description"],
            "meta": service_arguments["meta"],
        }
        return Result(json.dumps(output))

    async def getset(self, session, key, value) -> Result:
        if session.current_db < 0 or session.current_db >= self.MAX_DB_COUNT:
            return Result(DatabaseNotFound())
        current_service = self.SERVICE_MAP.get(session.current_db, None)
        if not current_service:
            return Result(DatabaseNotFound())
        service_map = current_service["map"]
        service_dict = service_map.get(key, None)
        if not service_dict:
            return Result(KeyNotFound())
        handler = service_dict["handler"]
        service_arguments = service_dict["args"]
        loads = current_service["loads"]
        dumps = current_service["dumps"]
        if not service_arguments.get("output", False):
            return Result(WrongCommand())
        try:
            signature = service_arguments["signature"]
            message = loads(value)
            if "kwargs" in signature.parameters:
                result = await handler(**message)
            else:
                result = await handler(message=message)
            return Result(dumps(result))
        except Exception as e:
            return Result(e)

    async def set(self, session, key, value) -> Result:
        if session.current_db < 0 or session.current_db >= self.MAX_DB_COUNT:
            return Result(DatabaseNotFound())
        current_service = self.SERVICE_MAP.get(session.current_db, None)
        if not current_service:
            return Result(DatabaseNotFound())
        service_map = current_service["map"]
        service_dict = service_map.get(key, None)
        if not service_dict:
            return Result(KeyNotFound())
        handler = service_dict["handler"]
        service_arguments = service_dict["args"]
        loads = current_service["loads"]
        try:
            signature = service_arguments["signature"]
            message = loads(value)
            if "kwargs" in signature.parameters:
                await handler(**message)
            else:
                await handler(message=message)
            return Result(True)
        except Exception as e:
            return Result(e)

    async def key_type(self, session, key):
        return Result("string")

    async def ping(self, session):
        return Result("PONG")

    def info_arguments(self):
        service_list = list()
        for db in range(self.MAX_DB_COUNT):
            service_list.append(
                (db,
                 dict(
                     keys=len(self.SERVICE_MAP.get(db, dict()).get("map", list())),
                 )
                 ))
        arguments = super(ServiceBasedRedisServer, self).info_arguments()
        arguments.update(porkpepper_mode="service", keyspace=service_list)
        return arguments

    async def info(self, session):
        info_arguments = self.info_arguments()
        info = self.INFO_TEMPLATE.render(info_arguments)
        return Result(info)

    async def ttl(self, session, key):
        return Result(-1)

    async def scan(self, session, pattern):
        if session.current_db < 0 or session.current_db >= self.MAX_DB_COUNT:
            return Result(DatabaseNotFound())
        current_service = self.SERVICE_MAP.get(session.current_db, None)
        if not current_service:
            return Result(list())
        service_map = current_service["map"]
        return Result(list(service_map.keys()))

    async def dbsize(self, session):
        if session.current_db < 0 or session.current_db >= self.MAX_DB_COUNT:
            return Result(DatabaseNotFound())
        current_service = self.SERVICE_MAP.get(session.current_db, None)
        if not current_service:
            return Result(0)
        service_map = current_service["map"]
        return Result(len(service_map))

    async def select(self, session, db):
        if 0 <= db < self.MAX_DB_COUNT:
            session.current_db = db
            return Result(True)
        else:
            return Result(DatabaseNotFound())

    async def config(self, session, get_set, field, value=None):
        if get_set == b'GET' and field == b'databases':
            return Result([b'databases', f'{self.MAX_DB_COUNT}'.encode("utf8")])
        return Result(CommandNotFound())


class RedisServiceNode(PorkPepperNode):
    def __init__(self, redis_server=ServiceBasedRedisServer, **kwargs):
        super(RedisServiceNode, self).__init__(redis_server=redis_server, **kwargs)

    def clear_service(self):
        if self._redis_server:
            self._redis_server.MAX_DB_COUNT = 0
            self._redis_server.SERVICE_MAP.clear()

    def update_service(self, db: int, new_service: Type):
        members = inspect.getmembers(new_service, inspect.ismethod)
        service_map = dict()
        for _, member in members:
            if not inspect.iscoroutinefunction(member):
                continue
            sign = inspect.signature(member)
            if "message" not in sign.parameters and "kwargs" not in sign.parameters:
                continue
            if not hasattr(member, PORKPEPPER_ATTR):
                continue
            service_arguments = getattr(member, PORKPEPPER_ATTR)
            key = service_arguments.get("key", None)
            if key in service_map:
                raise KeyError
            if not isinstance(key, str):
                raise ValueError
            service_map[key] = dict(handler=member, args=service_arguments)
        loads = getattr(new_service, "LOADS", json.loads)
        dumps = getattr(new_service, "DUMPS", json.dumps)
        if self._redis_server:
            if db + 1 > self._redis_server.MAX_DB_COUNT:
                self._redis_server.MAX_DB_COUNT = db + 1
            self._redis_server.SERVICE_MAP[db] = dict(loads=loads, dumps=dumps, map=service_map)
            
    async def _initialize_service_map(self, service_map: Dict[int, Type] = None):
        if service_map:
            for db, service_type in service_map.items():
                if not isinstance(db, int):
                    raise ValueError
                self.update_service(db, service_type)
        else:
            self.clear_service()

    async def start(self, service_map: Dict[int, Type] = None, redis_host="127.0.0.1", redis_port=6379, **kwargs):
        await self._initialize_service_map(service_map)
        return await super(RedisServiceNode, self).start(
            enable_websocket=False,
            redis_host=redis_host,
            redis_port=redis_port, **kwargs
        )

    async def serve(self, service_map: Dict[int, Type] = None, redis_host="127.0.0.1", redis_port=6379, **kwargs):
        await self._initialize_service_map(service_map)
        await super(RedisServiceNode, self).serve(
            enable_websocket=False,
            redis_host=redis_host,
            redis_port=redis_port,
            **kwargs
        )


__all__ = ["WebsocketNode", "RedisServiceNode", "service", "SocketBasedRedisServer", "ServiceBasedRedisServer", ]



from typing import *
from struct import pack
import time
import base58
import hashlib


class HashCipher(object):
    def __init__(self, salt=None, cipher=hashlib.blake2b):
        self._salt: Optional[bytes] = salt or b''
        self._cipher = cipher

    def hash(self, text) -> str:
        cipher = self._cipher(salt=self._salt)
        cipher.update(text.encode("utf8"))
        return cipher.hexdigest()

    def digest(self, text) -> bytes:
        cipher = self._cipher(salt=self._salt)
        cipher.update(text.encode("utf8"))
        return cipher.digest()


def create_base58_key(
        data,
        length: int = 16,
        prefix: str = "",
        salt: Optional[str]= None,
        cipher=hashlib.blake2b,
        timestamp: bool = False,
) -> str:
    hash_cipher = HashCipher(salt=salt, cipher=cipher)
    hash_key = hash_cipher.digest(str(data))
    slice_key = base58.b58encode(hash_key)
    if type(slice_key) is bytes:
        slice_key = slice_key.decode("utf8")
    time_part = ''
    if timestamp:
        now_timestamp = int(time.time() * 1000)
        time_part = base58.b58encode(pack('>Q', now_timestamp)).zfill(10).decode("utf8")
    key = f'{prefix}{"{}{}".format(time_part, slice_key)[:length]}'
    return key


__all__ = ["create_base58_key", ]



from typing import *
import asyncio
from aiohttp import web
from .websocket_app import WebsocketApp
from .redis_server_base import RedisServerBase


class PorkPepperNode:
    def __init__(self, redis_server, session_class=None, websocket_path="/porkpepper", **kwargs):
        if redis_server is None:
            raise ValueError
        self._session_class = session_class

        self._app_options = kwargs or dict()
        self._http_app: WebsocketApp = WebsocketApp(session_class=session_class, **self._app_options)
        self._http_app.add_routes([web.route('GET', path, self._http_app.handler) for path in [websocket_path, ]])

        self._redis_server_class = redis_server
        self._redis_server = self._redis_server_class(app=self._http_app)

        self._redis_server_task = None
        self._redis_server_object = None
        self._runner = None

        self.start_event = asyncio.Event()
        self.stop_event = asyncio.Event()

    async def on_start(self):
        pass

    async def on_shutdown(self):
        pass

    async def start(self, enable_websocket=False, redis_host="127.0.0.1", redis_port=6379, **kwargs):
        app = self._http_app
        redis_server: RedisServerBase = self._redis_server
        redis_server.init_property()
        runner = web.AppRunner(app)
        self._runner = runner
        await runner.setup()
        if enable_websocket:
            site = web.TCPSite(runner, **kwargs)
            await site.start()
        await self.on_start()
        redis_server_object = await redis_server.serve(redis_host, redis_port, False)
        self._redis_server_object = redis_server_object
        return self

    async def stop(self):
        if self._redis_server_object:
            self._redis_server_object.close()
            await self._redis_server_object.wait_closed()
            self._redis_server_object = None
        if self._redis_server_task:
            self._redis_server_task.cancel()
            self._redis_server_task = None
        if self._runner:
            await self._runner.cleanup()
            self._runner = None
        await self.on_shutdown()

    async def serve(self, enable_websocket=False, redis_host="127.0.0.1", redis_port=6379, **kwargs):
        self.start_event.clear()
        self.stop_event.clear()
        app = self._http_app
        redis_server: RedisServerBase = self._redis_server
        redis_server.init_property()
        runner = web.AppRunner(app)
        await runner.setup()
        try:
            if enable_websocket:
                site = web.TCPSite(runner, **kwargs)
                await site.start()
            await self.on_start()
            await redis_server.serve(redis_host, redis_port, True, self.start_event.set)
        finally:
            await runner.cleanup()
            await self.on_shutdown()
            self.stop_event.set()
